from .config import (get_kooki_dir, get_kooki_dir_jars, get_kooki_jar_config,
                     get_kooki_jar_manager)
from .parse import parse_document_rules

__all__ = [
    get_kooki_dir,
    get_kooki_jar_manager,
    get_kooki_dir_jars,
    get_kooki_jar_config,
    parse_document_rules,
]



import argparse
import sys

from karamel.command import (Command, InstallCommand, ListCommand,
                             SearchCommand, UninstallCommand, UpdateCommand)
from kooki.cli import BakeCommand
from kooki.config import get_kooki_dir_jars, get_kooki_jar_manager
from kooki.version import __version__

__program__ = "kooki"
__description__ = "Generate any kind of documents."

class KookiCommand(Command):
    def __init__(self):
        super(KookiCommand, self).__init__(__program__, __description__)
        self.add_argument(
            "-v",
            "--version",
            help="Show program's version number and exit.",
            action="store_true",
        )
        self.add_argument(
            "-h",
            "--help",
            action="help",
            default=argparse.SUPPRESS,
            help="Show this help message and exit.",
        )

        self.add_command(BakeCommand())

        package_manager_url = get_kooki_jar_manager()
        package_install_dir = get_kooki_dir_jars()

        self.add_command(SearchCommand(package_manager_url))
        self.add_command(InstallCommand(package_manager_url, package_install_dir))
        self.add_command(UninstallCommand(package_manager_url, package_install_dir))
        self.add_command(UpdateCommand(package_manager_url, package_install_dir))
        self.add_command(ListCommand(package_manager_url, package_install_dir))

    def callback(self, args):
        show_version(args)


def show_version(args):
    if args.version:
        print("{0} {1}\nPython {2}".format(__program__, __version__, sys.version))
        raise SystemExit(0)


def run():
    try:
        command = KookiCommand()
        command.run()

    except (KeyboardInterrupt, SystemExit):
        pass



import glob
import os
from functools import lru_cache

from packaging import version

from kooki.config import get_kooki_dir_jars, get_kooki_jar_config

from ..version import __version__
from .exception import KookiJarOutdated, TooMuchResult


def check_kooki_jar_version(jar_name):
    jar_config = get_kooki_jar_config(jar_name)
    if jar_config and "kooki" in jar_config:
        minimum_kooki_version_needed = jar_config["kooki"]
        if version.parse(__version__) < version.parse(
                minimum_kooki_version_needed):
            raise KookiJarOutdated(jar_name, __version__,
                                   minimum_kooki_version_needed)


def jar_dependencies(jar_name):
    jar_config = get_kooki_jar_config(jar_name)
    if jar_config and "dependencies" in jar_config:
        return jar_config["dependencies"]
    else:
        return []


def search_jar(jar):
    if "==" in jar:
        jar = jar.split("==")[0]
    ret_jar_path = None
    user_jars_dir = get_kooki_dir_jars()
    jar_path = os.path.join(user_jars_dir, jar)
    if os.path.isdir(jar_path):
        ret_jar_path = jar_path
    return ret_jar_path


def get_search_paths(jars):
    directories = []
    directories.append(os.getcwd())
    directories += get_jars_path(jars)
    return directories


def get_jars_path(jars):
    user_jars_dir = get_kooki_dir_jars()
    directories = []
    for jar in jars:
        if "==" in jar:
            jar = jar.split("==")[0]
        jar_path = os.path.join(user_jars_dir, jar)
        resource_jar_path = os.path.join(jar_path)
        directories.append(resource_jar_path)
    return directories


@lru_cache(maxsize=256)
def search_file_in_context(document, extension_path, path):
    context = os.path.join(os.getcwd(), document.context)
    directory_path = os.path.dirname(extension_path)
    search_path = os.path.join(os.path.join(directory_path, path))
    if search_path.startswith(context):
        return search_file_in_path(search_path)
    else:
        return None


@lru_cache(maxsize=256)
def search_file(document, filename):
    ret_file_path = None
    ret_file_path = search_file_in_local(
        os.path.join(document.context, filename))
    jars = list(reversed(document.jars))
    if not ret_file_path:
        ret_file_path = search_file_in_jars(jars, filename)
    return ret_file_path


def search_file_in_local(filename):
    ret_file_path = None
    file_path = os.path.join(os.getcwd(), filename)
    ret_file_path = search_file_in_path(file_path)
    return ret_file_path


def search_file_in_jars(jars, filename):
    user_jars_dir = get_kooki_dir_jars()
    ret_file_path = None
    for jar in jars:
        if "==" in jar:
            jar = jar.split("==")[0]
        jar_path = os.path.join(user_jars_dir, jar)
        ret_file_path = search_file_in_jar(jar_path, filename)
        if ret_file_path:
            break
    return ret_file_path


def search_file_in_jar(jar_path, filename):
    ret_file_path = None
    file_path = os.path.join(jar_path, filename)
    ret_file_path = search_file_in_path(file_path)
    return ret_file_path


def search_file_in_path(file_path):
    if os.path.isdir(file_path):
        return file_path
    else:
        root, ext = os.path.splitext(file_path)
        result = glob.glob("{}.*".format(file_path))

        if ext == "":
            result = glob.glob("{}.*".format(file_path))
            if len(result) == 1:
                return result[0]
            elif len(result) > 1:
                raise TooMuchResult(file_path, result)
            return None
        else:
            if os.path.isfile(file_path):
                return file_path
            else:
                return None



import argparse
import textwrap
import traceback

import pretty_output
from karamel.command import Command
from karamel.exception import KaramelException
from karamel.packages import install_packages
from kooki.config import (get_kooki_dir_jars, get_kooki_jar_manager,
                          parse_document_rules)
from kooki.exception import KookiException
from kooki.processor import process_document
from kooki.processor.jars import check_kooki_jar_version, jar_dependencies
from kooki.tools import read_config_file

__command__ = "bake"
__description__ = "Bake a kooki"


class BakeCommand(Command):
    def __init__(self):
        super(BakeCommand, self).__init__(__command__, __description__)

        help_help_message = "Show this help message and exit."
        debug_help_message = "Show information to help debug the bake processing"

        self.add_argument("documents", nargs="*")
        self.add_argument("--config-file", default="kooki.yaml")
        self.add_argument(
            "-h",
            "--help",
            action="help",
            default=argparse.SUPPRESS,
            help=help_help_message,
        )
        self.add_argument(
            "-d", "--debug", help=debug_help_message, action="store_true")
        self.add_argument(
            "--no-color", help="The output has no color.", action="store_true")
        self.add_argument(
            "--no-output", help="There is no output", action="store_true")
        self.add_argument("--update", help="Update jars.", action="store_true")

    def callback(self, args):
        try:
            pretty_output.set_output_policy(not args.no_output)
            pretty_output.set_color_policy(not args.no_color)
            pretty_output.set_debug_policy(args.debug)

            real_call(args)

        except KookiException as e:
            pretty_output.error_step("Errors")
            pretty_output.error(e)
            pretty_output.debug_on()
            pretty_output.error(traceback.format_exc()[:-1])
            pretty_output.debug_off()

        except Exception:
            pretty_output.error_step("Errors")
            pretty_output.error(traceback.format_exc()[:-1])


def real_call(args):
    config = read_config_file(args.config_file, must_exist=True)
    document_rules = parse_document_rules(config)
    do_update = args.update

    document = get_documents_to_generate(args, document_rules)

    if isinstance(document, dict):
        generate_documents(document, do_update)
    else:
        generate_document(document, do_update)


def get_documents_to_generate(args, document_rules):
    if args.documents == []:
        documents = document_rules
    else:
        documents = {}
        for document_name in args.documents:
            if document_name in document_rules:
                documents[document_name] = document_rules[document_name]
            else:
                raise Exception("Bad document")
    return documents


def generate_documents(documents, do_update):
    for name, document in documents.items():
        pretty_output.title_1(name)
        generate_document(document, do_update)
        pretty_output.title_2()


def generate_document(document, do_update):
    if do_update:
        execute_update(document)
    execute_check_kooki_version(document)
    execute_bake(document)


def on_package_downloading(package_name):
    print("Downloading '{}'".format(package_name))


def on_package_installing(package_name):
    print("Installing '{}'".format(package_name))


def on_package_install_success(package_name):
    print("Successfully installed '{}'".format(package_name))


def on_package_already_installed(package):
    def callback(package_name, package_path):
        message = textwrap.dedent("""\
            {} '{}' already installed in '{}'.
            """).format(package, package_name, package_path)
        print(message)

    return callback


def on_package_not_found(package):
    def callback(package_name):
        raise KookiException("{} not found '{}'.".format(
            package, package_name))

    return callback


def on_package_bad_version_provided(package):
    def callback(package_name, version):
        message = textwrap.dedent("""\
            Could not find the version '{1}' for {2} '{0}'.
            """).format(package_name, version, package)
        raise KookiException(message)

    return callback


def on_package_could_not_be_download(package):
    def callback(package_name):
        message = textwrap.dedent("""\
            {} \'{}\' could not be download.
            Without this package the document cannot be generated.
            Are you connected to the Internet ?""").format(package_name)
        raise KookiException(message.format(package, package_name))

    return callback


def execute_update(document):
    pretty_output.title_2("update")
    try:
        pretty_output.title_3("jars")
        install_jars(document.jars)
    except KaramelException as e:
        print(e)


def install_jars(jars):
    package_manager_url = get_kooki_jar_manager()
    package_install_dir = get_kooki_dir_jars()

    install_packages(
        package_manager_url,
        package_install_dir,
        jars,
        on_package_downloading,
        on_package_installing,
        on_package_install_success,
        on_package_already_installed("Jar"),
        on_package_not_found("Jar"),
        on_package_bad_version_provided("jar"),
        on_package_could_not_be_download("Jar"),
    )

    for jar in jars:
        jars = jar_dependencies(jar)
        install_jars(jars)


def execute_check_kooki_version(document):
    check_jar_kooki_version(document.jars)


def check_jar_kooki_version(jars):
    for jar in jars:
        check_kooki_jar_version(jar)
        jars = jar_dependencies(jar)
        check_jar_kooki_version(jars)


def execute_bake(document):
    pretty_output.title_2("bake")
    process_document(document, pretty_output.title_3)



import os

import yaml
from packaging import version

import pretty_output
from kooki.tools import read_file

from ..version import __version__
from .common import export_to
from .exception import (KookiOutdated, MissingJar, MissingMetadata,
                        MissingTemplate)
from .extension import Caller, Extension
from .jars import jar_dependencies, search_file, search_jar
from .metadata import parse_metadata
from .renderer import KookiRenderer

output = []


def output_search_start():
    global output
    output = []


def output_search(path, fullpath):
    if fullpath:
        output.append({"name": path, "status": "[found]", "path": fullpath})
    else:
        output.append({
            "name": path,
            "status": ("[missing]", "red"),
            "path": ""
        })


def output_search_finish():
    pretty_output.infos(output, [("name", "blue"), ("status", "green"),
                                 ("path", "cyan")])


def process_document(document, step=lambda *args, **kwargs: None):
    # clean
    KookiRenderer.clean()

    # jars
    step("jars")
    load_jars(document)

    # version
    check_kooki_version(document.version)

    # template
    step("template")
    template_path, template_extension = load_template(document)
    document.extension = template_extension
    template = Extension(document, template_path)

    # metadata
    step("metadata")
    load_metadata(document)

    # process
    step("processor")
    document_content = template()
    KookiRenderer.clean_double_pass()
    document_content = template()

    output = Extension(document, "output.tex", document.output)()
    # output = apply_template(document.output, document.metadata)
    output_path = os.path.join(document.context, output.replace("\n", ""))
    Caller.clean()

    # output
    step("output")
    absolute_file_path = export_to(output_path, template_extension,
                                   document_content)
    pretty_output._print_colored("export to ", "blue", "")
    pretty_output._print_colored(absolute_file_path, "cyan")

    # toppings
    step("toppings")
    import subprocess

    environ_variables = {
        "KOOKI_CONTEXT": document.context,
        "KOOKI_OUTPUT": output_path,
        "KOOKI_EXPORT_FILE": absolute_file_path,
    }

    output = []
    for name, value in environ_variables.items():
        output.append({"name": name, "value": value})
        os.environ[name] = value
    pretty_output.infos(output, [("name", "blue"), ("value", "cyan")])

    for topping in document.toppings:
        pretty_output._print_colored("execute ", "blue", "")
        pretty_output._print_colored(topping, "cyan")
        subprocess.call(topping, shell=True)
        pretty_output._print_colored("finish ", "blue", "")
        pretty_output._print_colored(topping, "cyan")


def check_kooki_version(kooki_version_needed):
    if version.parse(__version__) < version.parse(kooki_version_needed):
        raise KookiOutdated(__version__, kooki_version_needed)


def load_template(document):
    output_search_start()
    file_full_path = search_file(document, document.template)
    output_search(document.template, file_full_path)
    if file_full_path:
        file_read = read_file(file_full_path)
        document.template = file_read
        output_search_finish()
    else:
        output_search_finish()
        raise MissingTemplate(document.jars, document.template)

    _, template_extension = os.path.splitext(file_full_path)
    return file_full_path, template_extension


def print_extension(extensions, prefix=""):
    extensions_name = list(extensions.keys())
    extensions_name.sort()
    if prefix != "":
        prefix += "."
    for extension_name in extensions_name:
        extension = extensions[extension_name]
        if isinstance(extension, dict):
            print_extension(extension, extension_name)
        else:
            output_search("{}{}".format(prefix, extension_name),
                          extension.path)


def load_jars(document):
    full_path_jars = []
    output_search_start()
    for jar in document.jars:
        load_jar_dependencie(full_path_jars, jar)
        jar_full_path = search_jar(jar)
        if jar_full_path:
            output_search(jar, jar_full_path)
            full_path_jars.append(jar_full_path)
        else:
            output_search_finish()
            raise MissingJar(jar)
    document.jars = full_path_jars
    output_search_finish()


def load_jar_dependencie(full_path_jars, jar):
    dependencies = jar_dependencies(jar)
    for dependencie in dependencies:
        load_jar_dependencie(full_path_jars, dependencie)
        jar_full_path = search_jar(dependencie)
        if jar_full_path:
            output_search(dependencie, jar_full_path)
            full_path_jars.append(jar_full_path)
        else:
            output_search_finish()
            raise MissingJar(jar)


def load_metadata(document):
    metadata_full_path = {}
    output_search_start()
    for metadata in document.metadata:
        file_full_path = search_file(document, metadata)
        output_search(metadata, file_full_path)
        if file_full_path:
            file_read = read_file(file_full_path)
            metadata_full_path[file_full_path] = file_read
        else:
            output_search_finish()
            raise MissingMetadata(document.jars, metadata)
    document.metadata = metadata_full_path
    output_search_finish()
    document.metadata = parse_metadata(document.metadata)

    pretty_output.debug_on()
    pretty_output.colored(
        yaml.dump(document.metadata, default_flow_style=False), "white",
        "on_yellow")
    pretty_output.debug_off()


def print_metadata(metadata, prefix=""):
    matadata_keys = list(metadata.keys())
    matadata_keys.sort()
    if prefix != "":
        prefix += "."
    for matadata_key in matadata_keys:
        element = metadata[matadata_key]
        if isinstance(element, dict):
            print_metadata(element, matadata_key)
        else:
            output_search("{}{}".format(prefix, matadata_key), "")



import os

import yaml

from kooki.tools import read_config_file

resource_dir_env = "KOOKI_DIR"
resource_dir_default = "~/.kooki"

jar_manager_env = "KOOKI_JAR_MANAGER"
jar_manager_default = [
    "https://gitlab.com/kooki/jar_manager/raw/master/jars.yml"
]

jar_config_file = "jar.yaml"


def get_kooki_dir():
    resource_dir = os.environ.get(resource_dir_env)
    if not resource_dir:
        resource_dir = os.path.expanduser(resource_dir_default)
    return resource_dir


def get_kooki_dir_jars():
    resources_dir = get_kooki_dir()
    jars_dir = os.path.join(resources_dir, "jars")
    return jars_dir


def get_kooki_jar_manager():
    jar_manager = os.environ.get(jar_manager_env)
    if jar_manager:
        jar_manager = yaml.safe_load(jar_manager)
    else:
        jar_manager = jar_manager_default
    return jar_manager


def get_kooki_jar_config(jar_name):
    user_jars_dir = get_kooki_dir_jars()
    jar_config_file_path = os.path.join(user_jars_dir, jar_name,
                                        jar_config_file)
    jar_config = read_config_file(jar_config_file_path, must_exist=False)
    return jar_config



from cerberus import Validator

from .document import Document

document_format = Document.format()
document_fields = Document.fields()

document_version_1 = {
    **document_format,
    "default": {
        "required": False,
        "type": "dict",
        "excludes": document_fields,
        "dependencies": "kooki",
        "schema": document_format,
    },
    "kooki": {
        "required": False,
        "type": "dict",
        "excludes": document_fields,
        "keyschema": {
            "type": "string",
            "regex": "[A-Za-z0-9_-]+"
        },
    },
}


def check_config_format(config):
    v = Validator(document_version_1)
    config_valid = v.validate(config, document_version_1)
    if not config_valid:
        raise Exception(v.errors)



import re
import uuid

from mistune import InlineGrammar, InlineLexer, Markdown, Renderer

from munch import Munch


def render(content, extensions):
    renderer = KookiRenderer(extensions)
    markdown = KookiMarkdown(renderer=renderer)
    return markdown(content)


class KookiInlineGrammar(InlineGrammar):

    criticmarkup_deletion = re.compile(r"^{--(?=\S)([\s\S]*?\S)--}")
    criticmarkup_substitution = re.compile(
        r"^{~~(?=\S)([\s\S]*?\S)~>(?=\S)([\s\S]*?\S)~~}")
    criticmarkup_comment = re.compile(r"^{>>(?=\S)([\s\S]*?\S)<<}")
    criticmarkup_highlight = re.compile(
        r"^{==(?=\S)([\s\S]*?\S)==}{>>(?=\S)([\s\S]*?\S)<<}")
    criticmarkup_addition = re.compile(r"^{\+\+(?=\S)([\s\S]*?\S)\+\+}")
    text = re.compile(r"^[\s\S]+?(?=[\\<!\[_*`~{}]|https?://| {2,}\n|$)")


class KookiInlineLexer(InlineLexer):

    default_rules = [
        "escape",
        "inline_html",
        "autolink",
        "url",
        "footnote",
        "link",
        "reflink",
        "nolink",
        "double_emphasis",
        "emphasis",
        "code",
        "criticmarkup_addition",
        "criticmarkup_deletion",
        "criticmarkup_substitution",
        "criticmarkup_comment",
        "criticmarkup_highlight",
        "linebreak",
        "strikethrough",
        "text",
    ]

    def __init__(self, renderer, **kwargs):
        rules = KookiInlineGrammar()
        super(KookiInlineLexer, self).__init__(renderer, rules, **kwargs)

    def output_criticmarkup_addition(self, m):
        text = m.group(1)
        return self.renderer.criticmarkup_addition(text)

    def output_criticmarkup_deletion(self, m):
        text = m.group(1)
        return self.renderer.criticmarkup_deletion(text)

    def output_criticmarkup_substitution(self, m):
        text = m.group(1)
        substitution = m.group(2)
        return self.renderer.criticmarkup_substitution(text, substitution)

    def output_criticmarkup_comment(self, m):
        text = m.group(1)
        return self.renderer.criticmarkup_comment(text)

    def output_criticmarkup_highlight(self, m):
        text = m.group(1)
        comment = m.group(2)
        return self.renderer.criticmarkup_highlight(text, comment)


class KookiMarkdown(Markdown):
    def __init__(self, renderer, **kwargs):
        if "inline" not in kwargs:
            kwargs["inline"] = KookiInlineLexer(renderer)
        super(KookiMarkdown, self).__init__(renderer, **kwargs)


def catch_errors(func):
    def callback(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            return ""

    return callback


class TocItem:
    def __init__(self, title, uid):
        self.title = title
        self.uid = uid
        self.children = []


class KookiRenderer(Renderer):

    data = Munch()
    toc = []
    toc_h2 = []
    toc_h3 = []
    toc_h4 = []
    toc_h5 = []
    toc_h6 = []
    toc_history = {}

    @classmethod
    def clean(cls):
        cls.data = Munch()
        cls.toc = []
        cls.toc_h2 = []
        cls.toc_h3 = []
        cls.toc_h4 = []
        cls.toc_h5 = []
        cls.toc_h6 = []
        cls.toc_history = {}

    @classmethod
    def clean_double_pass(cls):
        cls.data = Munch()

    def __init__(self, extensions, **kwargs):
        self.extensions = extensions
        super(KookiRenderer, self).__init__(**kwargs)

    def block_quote(self, text):
        return self.extensions.__block_quote__(
            renderer=KookiRenderer.data, text=text)

    def block_html(self, html):
        return self.extensions.__block_html__(
            renderer=KookiRenderer.data, html=html)

    def inline_html(self, html):
        return self.extensions.__inline_html__(
            renderer=KookiRenderer.data, html=html)

    def header(self, text, level, raw):

        if raw in KookiRenderer.toc_history:
            uid = KookiRenderer.toc_history[raw].uid
            return self.extensions.__header__(
                renderer=KookiRenderer.data,
                text=text,
                level=level,
                raw=raw,
                uid=uid)

        else:
            uid = uuid.uuid4()
            KookiRenderer.toc_history[raw] = TocItem(text, uid)
            if level == 1:
                if len(KookiRenderer.toc_h2) > 0:
                    KookiRenderer.toc[-1].children = KookiRenderer.toc_h2
                    KookiRenderer.toc_h2 = []
                KookiRenderer.toc.append(TocItem(text, uid))
            elif level == 2:
                if len(KookiRenderer.toc_h3) > 0:
                    KookiRenderer.toc_h2[-1].children = KookiRenderer.toc_h3
                    KookiRenderer.toc_h3 = []
                KookiRenderer.toc_h2.append(TocItem(text, uid))
            elif level == 3:
                if len(KookiRenderer.toc_h4) > 0:
                    KookiRenderer.toc_h3[-1].children = KookiRenderer.toc_h4
                    KookiRenderer.toc_h4 = []
                KookiRenderer.toc_h3.append(TocItem(text, uid))
            elif level == 4:
                if len(KookiRenderer.toc_h5) > 0:
                    KookiRenderer.toc_h4[-1].children = KookiRenderer.toc_h5
                    KookiRenderer.toc_h5 = []
                KookiRenderer.toc_h4.append(TocItem(text, uid))
            elif level == 5:
                if len(KookiRenderer.toc_h6) > 0:
                    KookiRenderer.toc_h5[-1].children = KookiRenderer.toc_h6
                    KookiRenderer.toc_h6 = []
                KookiRenderer.toc_h5.append(TocItem(text, uid))
            elif level == 6:
                KookiRenderer.toc_h6.append(TocItem(text, uid))

            return self.extensions.__header__(
                renderer=KookiRenderer.data,
                text=text,
                level=level,
                raw=raw,
                uid=uid)

    def hrule(self):
        return self.extensions.__hrule__(renderer=KookiRenderer.data)

    def list(self, body, ordered):
        return self.extensions.__list__(
            renderer=KookiRenderer.data, body=body, ordered=ordered)

    def list_item(self, text):
        return self.extensions.__list_item__(
            renderer=KookiRenderer.data, text=text)

    def paragraph(self, text):
        return "{}\n\n".format(
            self.extensions.__paragraph__(
                renderer=KookiRenderer.data, text=text))

    def table(self, header, body):
        return self.extensions.__table__(
            renderer=KookiRenderer.data, header=header, body=body)

    def table_row(self, content):
        return self.extensions.__table_row__(
            renderer=KookiRenderer.data, content=content, placeholder=False)

    def table_cell(self, content, **flags):
        return self.extensions.__table_cell__(
            renderer=KookiRenderer.data,
            content=content,
            placeholder=False,
            **flags)

    def link(self, link, title, content):
        return self.extensions.__link__(
            renderer=KookiRenderer.data,
            link=link,
            title=title,
            content=content)

    def autolink(self, link, is_email=False):
        return self.extensions.__autolink__(
            renderer=KookiRenderer.data, link=link, is_email=is_email)

    def block_code(self, code, language):
        return self.extensions.__block_code__(
            renderer=KookiRenderer.data, code=code, language=language)

    def codespan(self, text):
        return self.extensions.__codespan__(
            renderer=KookiRenderer.data, text=text)

    def double_emphasis(self, text):
        return self.extensions.__double_emphasis__(
            renderer=KookiRenderer.data, text=text)

    def emphasis(self, text):
        return self.extensions.__emphasis__(
            renderer=KookiRenderer.data, text=text)

    def image(self, src, title, alt_text):
        notitle = title is None
        return self.extensions.__image__(
            renderer=KookiRenderer.data,
            src=src,
            notitle=notitle,
            title=title,
            alt_text=alt_text,
        )

    def strikethrough(self, text):
        return self.extensions.__strikethrough__(
            renderer=KookiRenderer.data, text=text)

    def text(self, text):
        return self.extensions.__text__(renderer=KookiRenderer.data, text=text)

    def linebreak(self):
        return self.extensions.__linebreak__(renderer=KookiRenderer.data)

    def newline(self):
        return self.extensions.__newline__(renderer=KookiRenderer.data)

    def footnote_ref(self, key, index):
        return self.extensions.__footnote_ref__(
            renderer=KookiRenderer.data, key=key, index=index)

    def footnote_item(self, key, text):
        return self.extensions.__footnote_item__(
            renderer=KookiRenderer.data, key=key, text=text)

    def footnotes(self, text):
        return self.extensions.__footnotes__(
            renderer=KookiRenderer.data, text=text)

    @catch_errors
    def criticmarkup_addition(self, text):
        return self.extensions.__criticmarkup_addition__(
            renderer=KookiRenderer.data, text=text)

    @catch_errors
    def criticmarkup_deletion(self, text):
        return self.extensions.__criticmarkup_deletion__(
            renderer=KookiRenderer.data, text=text)

    @catch_errors
    def criticmarkup_substitution(self, text, substitution):
        return self.extensions.__criticmarkup_substitution__(
            renderer=KookiRenderer.data, text=text, substitution=substitution)

    @catch_errors
    def criticmarkup_comment(self, text):
        return self.extensions.__criticmarkup_comment__(
            renderer=KookiRenderer.data, text=text)

    @catch_errors
    def criticmarkup_highlight(self, text, comment):
        return self.extensions.__criticmarkup_highlight__(
            renderer=KookiRenderer.data, text=text, comment=comment)



class String:
    def __init__(self, required=True, default=""):
        self.required = required
        self.default = default
        self.type = "string"

    def __call__(self):
        return self.default


class List:
    def __init__(self, required=True, default=[]):
        self.required = required
        self.default = default
        self.type = "list"

    def __call__(self):
        return self.default.copy()


class DocumentMetaclass(type):
    def __new__(cls, name, bases, dct):
        attrs = {}
        model = type.__new__(cls, name, bases, attrs)
        model.__init__ = cls.get_init(dct)
        model.format = cls.get_format(dct)
        model.fields = cls.get_fields(dct)
        return model

    @classmethod
    def get_format(cls, dct):
        @classmethod
        def callback(cls):
            rule_format = {}
            for attr_name, attr_value in dct.items():
                if isinstance(attr_value, (String, List)):
                    rule_format[attr_name] = {
                        "required": attr_value.required,
                        "type": attr_value.type,
                    }
            return rule_format

        return callback

    @classmethod
    def get_fields(cls, dct):
        @classmethod
        def callback(cls):
            return list(dct.keys())

        return callback

    @classmethod
    def get_init(cls, dct):
        def callback(self, *configs):
            for attr_name, attr_value in dct.items():
                if isinstance(attr_value, (String, List)):
                    setattr(self, attr_name, attr_value())

            for config in configs:
                for attr_name, attr_value in config.items():
                    if hasattr(self, attr_name):
                        if isinstance(attr_value, list):
                            old_attr = getattr(self, attr_name)
                            old_attr += attr_value
                            setattr(self, attr_name, old_attr)
                        elif isinstance(attr_value, str):
                            setattr(self, attr_name, attr_value)

        return callback


class Document(metaclass=DocumentMetaclass):
    version = String(required=False, default="0.0.0")
    output = String(required=False, default="output")
    name = String(required=False, default="name")
    template = String(required=False, default="")
    context = String(required=False, default=".")
    jars = List(required=False, default=[])
    metadata = List(required=False, default=[])
    toppings = List(required=False, default=[])



from .check import check_config_format
from .document import Document


def parse_document_rules(config):
    check_config_format(config)
    if "kooki" in config:
        return parse_multi_documents(config)
    else:
        return parse_mono_document(config)


def parse_mono_document(config):
    document = Document(config)
    return document


def parse_multi_documents(config):
    documents = {}
    for document_name, document_config in config["kooki"].items():

        if "default" in config:
            if document_config:
                document_config["name"] = document_name
                if ("output" not in config["default"]
                        and "output" not in document_config):
                    document_config["output"] = document_name
                if ("context" not in config["default"]
                        and "context" not in document_config):
                    document_config["context"] = document_name
            else:
                document_config = {}
                document_config["name"] = document_name
                document_config["output"] = document_name
                document_config["context"] = document_name
            document = Document(config["default"], document_config)

        else:
            if document_config:
                document_config["name"] = document_name
                if "output" not in document_config:
                    document_config["output"] = document_name
                if "context" not in document_config:
                    document_config["context"] = document_name
            else:
                document_config = {}
                document_config["name"] = document_name
                document_config["output"] = document_name
                document_config["context"] = document_name
            document = Document(document_config)

        documents[document_name] = document
    return documents



import os
import random
import string
import tempfile
import textwrap
from functools import lru_cache

import empy
import pretty_output
from kooki.tools import get_front_matter, read_file

from .exception import ErrorEvaluatingExpression, MissingFile
from .jars import search_file, search_file_in_context
from .metadata import Metadata
from .renderer import KookiRenderer, render

PREFIX = "@"
MARKDOWN_EXT = ".md"


class Extension:

    level = 0

    def __init__(self, document, extension_path, content=None):
        if content is not None:
            file_content = content
        else:
            file_content = read_file(extension_path)

        _, self.file_extension = os.path.splitext(extension_path)
        self.extension_path = extension_path
        self.document = document
        self.front_matter, self.content = get_front_matter(file_content)

    def __call__(self, *args, **kwargs):

        try:
            metadata = self.parse_metadata(*args, **kwargs)
            content = self.content

            self.debug_tree_start()

            caller = Caller(
                self.document,
                self.extension_path,
                metadata=metadata,
                file_extension=self.file_extension,
            )
            content = self.apply_interpreter(content, caller, prefix=PREFIX)

            if self.file_extension == MARKDOWN_EXT:
                content = render(content, caller)

            interpreter = empy.Interpreter()
            interpreter.setPrefix("§")
            content = interpreter.expand(content, Caller.placeholders)

            self.debug_tree_stop()

            return content

        except ErrorEvaluatingExpression as e:
            e.add_path(self.extension_path)
            raise e

    def apply_interpreter(self, content, metadata, prefix):
        interpreter = empy.Interpreter()
        interpreter.setPrefix(prefix)
        result = interpreter.expand(content, metadata)
        return result

    def parse_metadata(self, *args, **kwargs):
        metadata = Metadata()
        metadata.update(self.front_matter)
        metadata.update(self.document.metadata)
        new_args = {}
        for index, arg in enumerate(args):
            new_args["arg{}".format(index)] = arg
        metadata.update(**new_args)
        metadata.update(**kwargs)
        return metadata

    def debug_tree_start(self):
        pretty_output.debug_on()
        if Extension.level == 0:
            pretty_output.colored(self.extension_path, "white", "on_yellow")
        else:
            message = ""
            for i in range(0, self.level - 1):
                message += "│   "
            message += "├"
            message += "─" * 2
            message += " "
            pretty_output.colored(message, "white", "on_yellow", end="")
            pretty_output.colored(self.extension_path, "white", "on_yellow")
        Extension.level += 1
        pretty_output.debug_off()

    def debug_tree_stop(self):
        Extension.level -= 1


def is_builtin_class_instance(obj):
    return obj.__class__.__module__ == "__builtin__"


class Caller:

    placeholders = {}
    tempary_extensions = []

    @classmethod
    def clean(cls):
        for tempary_extension in cls.tempary_extensions:
            os.unlink(tempary_extension.name)

    def __init__(self,
                 document,
                 extension_path,
                 metadata,
                 file_extension,
                 call_path=""):
        self.document = document
        self.extension_path = extension_path
        self.call_path = call_path
        self.metadata = metadata
        self.file_extension = file_extension

        self.kooki_embeded_functions = Metadata()
        self.kooki_embeded_functions["kooki"] = Metadata({
            "find":
            self.find,
            "load":
            self.load,
            "ext":
            self.ext,
            "get":
            self.get,
            "name":
            self.document.name,
            "toc":
            KookiRenderer.toc,
        })

        self.metadata.update({
            "find": self.find_deprecated,
            "load": self.load_deprecated,
            "ext": self.ext_deprecated,
            "get": self.get_deprecated,
        })

        self.metadata.update(self.kooki_embeded_functions)
        self.data = None

    def __getattr__(self, name):
        return self.get_item(name)

    def __getitem__(self, key):
        return self.get_item(key)

    def __setitem__(self, key, value):
        self.metadata[key] = value

    def get_item(self, key):

        if key == "locals":

            def call_locals():
                return locals()

            return call_locals

        if self.call_path != "":
            new_call_path = "{}/{}".format(self.call_path, key)
        else:
            new_call_path = key

        self.metadata.update(__builtins__)
        self.metadata.update(locals()["self"].metadata)
        value = self.metadata.get(key)

        if value is not None:
            if is_builtin_class_instance(value):
                return type("", (type(value), Caller), {})(
                    self.document,
                    self.extension_path,
                    self.metadata,
                    file_extension=self.file_extension,
                    call_path=new_call_path,
                )
            else:
                return value
        else:
            return Caller(
                self.document,
                self.extension_path,
                self.metadata,
                file_extension=self.file_extension,
                call_path=new_call_path,
            )

    def __repr__(self):
        return ""

    def __call__(self, *args, placeholder=True, **kwargs):

        path = search_file_in_context(self.document, self.extension_path,
                                      self.call_path)
        if not path:
            path = search_file(self.document, self.call_path)

        if not path:
            raise Exception("cannot find {}".format(self.call_path))

        if os.path.isdir(path):
            full_path_md = os.path.join(path, "__kooki__.md")
            full_path_spec = os.path.join(
                path, "__kooki__{}".format(self.file_extension))
            if os.path.isfile(full_path_md):
                full_path = full_path_md
            elif os.path.isfile(full_path_spec):
                full_path = full_path_spec
            else:
                raise Exception(
                    textwrap.dedent("""\
                    You try to call this extension: {}
                    Cannot find one of those files in the directory.
                    - {}
                    - {}
                    """.format(path, full_path_md, full_path_spec)))
        else:
            full_path = path

        extension = self.create_extension(full_path)
        result = extension(*args, **kwargs)
        if placeholder:
            name = self.get_var_name()
            Caller.placeholders[name] = result
            return "§({})".format(name)
        else:
            return result

    def deprecation_message(self, message):
        pretty_output.warning(
            textwrap.dedent("""\
            Warning: {}
            {}
        """.format(self.extension_path, message)))

    @lru_cache(maxsize=256)
    def create_extension(self, full_path):
        extension = Extension(self.document, full_path)
        return extension

    def find_deprecated(self, file_path):
        self.deprecation_message(
            """'find' alone is deprecated, use 'kooki.find' instead.""")
        return self.find(file_path)

    def find(self, file_path):
        file_full_path = search_file(self.document, file_path)
        if file_full_path:
            return file_full_path
        else:
            raise MissingFile(self.document.jars, file_path)

    def ext_deprecated(self, path, *args, **kwargs):
        self.deprecation_message(
            """'ext' alone is deprecated, use 'kooki.ext' instead""")
        return self.ext(path, *args, **kwargs)

    def ext(self, path, *args, **kwargs):
        path = os.path.join(os.getcwd(), path)
        pretty_output.info(path)
        if os.path.isdir(path):
            full_path_md = os.path.join(path, "__kooki__.md")
            full_path_spec = os.path.join(
                path, "__kooki__{}".format(self.file_extension))
            if os.path.isfile(full_path_md):
                full_path = full_path_md
            elif os.path.isfile(full_path_spec):
                full_path = full_path_spec
            else:
                raise Exception(
                    textwrap.dedent("""\
                    You try to call this extension: {}
                    Cannot find one of those files in the directory.
                    - {}
                    - {}
                    """.format(path, full_path_md, full_path_spec)))
        else:
            full_path = path

        extension = self.create_extension(full_path)
        result = extension(*args, **kwargs)
        name = self.get_var_name()
        Caller.placeholders[name] = result
        return "§({})".format(name)

    def load_deprecated(self, content, template_extension):
        self.deprecation_message(
            """'load' alone is deprecated, use 'kooki.load' instead""")
        return self.load(content, template_extension)

    def load(self, content, template_extension):
        directory_path = os.path.dirname(self.extension_path)
        temp_extension = tempfile.NamedTemporaryFile(
            suffix=template_extension, dir=directory_path, delete=False)
        temp_extension.write(bytes(content, "utf-8"))
        temp_extension.close()
        Caller.tempary_extensions.append(temp_extension)
        return Extension(self.document, temp_extension.name)

    def get_deprecated(self, key, default=""):
        self.deprecation_message(
            """'get' alone is deprecated, use 'kooki.get' instead""")
        return self.get(key, default)

    def get(self, key, default=""):
        if key in self.metadata:
            return self.metadata[key]
        else:
            return default

    def get_var_name(self):
        name = "".join(
            random.choice(string.ascii_uppercase) for _ in range(25))
        while name in Caller.placeholders:
            name = "".join(
                random.choice(string.ascii_uppercase) for _ in range(25))
        return name
